"""
This module contains a class to store _all_ kinds of content evaluation results.
"""

from typing import Dict, Optional
from uuid import UUID

import attrs
from marshmallow import Schema, fields, post_load

from ahbicht.expressions.condition_nodes import (
    ConditionFulfilledValue,
    EvaluatedFormatConstraint,
    EvaluatedFormatConstraintSchema,
)


# pylint: disable=too-few-public-methods, no-self-use, unused-argument
@attrs.define(auto_attribs=True)
class ContentEvaluationResult:
    """
    A class that holds the results of a full content evaluation (meaning all hints, requirement constraints and
    format constraints have been evaluated)
    """

    #: maps the key of a hint (e.g. "501" to a hint text)
    hints: Dict[str, Optional[str]] = attrs.field(
        validator=attrs.validators.deep_mapping(
            key_validator=attrs.validators.instance_of(str),
            value_validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        )
    )

    #: maps the key of a format constraint to the respective evaluation result
    format_constraints: Dict[str, EvaluatedFormatConstraint] = attrs.field(
        validator=attrs.validators.deep_mapping(
            key_validator=attrs.validators.instance_of(str),
            value_validator=attrs.validators.instance_of(EvaluatedFormatConstraint),
        )
    )
    #: maps the key of a requirement_constraint to the respective evaluation result
    requirement_constraints: Dict[str, ConditionFulfilledValue] = attrs.field(
        validator=attrs.validators.deep_mapping(
            key_validator=attrs.validators.instance_of(str),
            value_validator=attrs.validators.instance_of(ConditionFulfilledValue),
        )
    )

    packages: Optional[Dict[str, str]] = attrs.field(  # Union[str, ConditionFulfilledValue]]
        validator=attrs.validators.optional(
            attrs.validators.deep_mapping(  # type: ignore[arg-type]
                key_validator=attrs.validators.and_(
                    attrs.validators.instance_of(str),
                    attrs.validators.matches_re(r"^\d+P$"),  # this is to avoid someone passes '123' instead of '123P'
                ),
                # todo: implement wiederholbarkeiten
                value_validator=attrs.validators.instance_of(str),
            )
        ),
        default=None,
    )
    """
    maps the key of a package (e.g. '123') to the respective expression (e.g. '[1] U ([2] O [3])'
    """

    # pylint:disable=invalid-name
    #: optional guid
    id: Optional[UUID] = attrs.field(
        validator=attrs.validators.optional(attrs.validators.instance_of(UUID)), default=None
    )


class ContentEvaluationResultSchema(Schema):
    """
    A schema to (de)serialize ContentEvaluationResults
    """

    hints = fields.Dict(keys=fields.String(allow_none=False), values=fields.String(allow_none=True), required=True)
    format_constraints = fields.Dict(
        keys=fields.String(allow_none=False),
        values=fields.Nested(EvaluatedFormatConstraintSchema, allow_none=False, required=True),
        required=True,
    )
    requirement_constraints = fields.Dict(
        keys=fields.String(allow_none=False), values=fields.String(allow_none=False), required=True
    )
    packages = fields.Dict(
        keys=fields.String(allow_none=False),
        values=fields.String(allow_none=False),
        required=False,
        load_default={},
        allow_none=True,
    )
    id = fields.UUID(required=False, dump_default=False, load_default=None)

    @post_load
    def deserialize(self, data, **kwargs) -> ContentEvaluationResult:
        """
        Converts the barely typed data dictionary into an actual ContentEvaluationResult
        """
        if "requirement_constraints" in data:
            # because of attrs.validators.deep_iterable this parsing step has to happen before binding to the instance
            for rc_key, rc_value in data["requirement_constraints"].items():
                if not isinstance(rc_value, ConditionFulfilledValue):
                    for enum_value in ConditionFulfilledValue:
                        if str(rc_value).upper() == enum_value.value:
                            data["requirement_constraints"][rc_key] = ConditionFulfilledValue(enum_value.value)
                            break
        result = ContentEvaluationResult(**data)
        return result



"""
Schemata for the JSON serialization of expressions.
"""
from lark import Tree
from marshmallow import post_dump, pre_load

from ahbicht.json_serialization.tree_schema import TreeSchema


class ConciseConditionKeyTreeSchema(TreeSchema):
    """
    A tree schema that merges condition key values and their single child
    (not deserialization)
    """

    # pylint:disable=unused-argument,no-self-use
    @post_dump
    def serialize(self, data, **kwargs) -> dict:
        """
        Serialize as compact/concise tree
        """
        return _compress_condition_keys_only(data)

    @pre_load
    def deserialize(self, data, **kwargs) -> Tree:
        """
        This is out of scope
        """
        raise NotImplementedError("Deserializing a compact tree is not supported.")


def _compress_condition_keys_only(data: dict) -> dict:
    """
    a function that merges a condition key node with its only child (a token that has an int value)
    """
    # this has been found heuristically. There's no way to explain it, just follow the test cases.
    # there's probably a much easier way, e.g. by using a separate token schema.
    if (
        "tree" in data
        and "type" in data["tree"]
        and data["tree"]["type"] == "single_requirement_indicator_expression"
        and data["tree"]["children"][0]["token"]["type"] == "MODAL_MARK"
    ):
        modal_mark = data["tree"]["children"][0]["token"]["value"]
        del data["tree"]["children"][0]
        data["tree"]["type"] = modal_mark
    if "tree" in data and data["tree"] is not None and data["tree"]["type"] == "condition":
        return {
            "token": {"value": data["tree"]["children"][0]["token"]["value"], "type": "condition_key"},
            "tree": None,
        }
    if (
        "token" in data
        and data["token"] is None
        and "tree" in data
        and data["tree"] is not None
        and "children" in data["tree"]
    ):
        data["tree"]["children"] = [_compress_condition_keys_only(child) for child in data["tree"]["children"]]
    if "type" in data and data["type"] is not None and "children" in data and data["children"] is not None:
        data["children"] = [_compress_condition_keys_only(child) for child in data["children"]]
    return data



# type:ignore[misc]
""" Tests for the parsing of the ahb_expressions as they appear in the AHBs. """

import pytest  # type:ignore[import]
from lark import Token, Tree

from ahbicht.expressions.ahb_expression_parser import parse_ahb_expression_to_single_requirement_indicator_expressions


class TestAhbExpressionParser:
    """Tests for the parsing of the ahb_expressions as they appear in the AHBs."""

    @pytest.mark.parametrize(
        "ahb_expression, expected_tree",
        [
            pytest.param(
                "Muss",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "requirement_indicator",
                            [
                                Token("MODAL_MARK", "Muss"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "X",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "requirement_indicator",
                            [
                                Token("PREFIX_OPERATOR", "X"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Muss[1]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Muss"),
                                Token("CONDITION_EXPRESSION", "[1]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Soll[1]U[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Soll"),
                                Token("CONDITION_EXPRESSION", "[1]U[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "soll[1]u[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "soll"),
                                Token("CONDITION_EXPRESSION", "[1]u[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Muss[UB1]U[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Muss"),
                                Token("CONDITION_EXPRESSION", "[UB1]U[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Kann([1]O[5])U[904]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Kann"),
                                Token("CONDITION_EXPRESSION", "([1]O[5])U[904]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Kann([1]∨[5])∧[904]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Kann"),
                                Token("CONDITION_EXPRESSION", "([1]∨[5])∧[904]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "X[1]O[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "X"),
                                Token("CONDITION_EXPRESSION", "[1]O[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "O[1]O[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "O"),
                                Token("CONDITION_EXPRESSION", "[1]O[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "O([1]U[5]) U\t[905]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "O"),
                                Token("CONDITION_EXPRESSION", "([1]U[5]) U\t[905]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Kann([1]U[5])U[905]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Kann"),
                                Token("CONDITION_EXPRESSION", "([1]U[5])U[905]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Kann([1]∧[5])∧[905]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Kann"),
                                Token("CONDITION_EXPRESSION", "([1]∧[5])∧[905]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "kann([1]∧[5])∧[905]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "kann"),
                                Token("CONDITION_EXPRESSION", "([1]∧[5])∧[905]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "Muss[3]U[4]Soll[5]    Kann[502]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Muss"),
                                Token("CONDITION_EXPRESSION", "[3]U[4]"),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Soll"),
                                Token("CONDITION_EXPRESSION", "[5]    "),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "Kann"),
                                Token("CONDITION_EXPRESSION", "[502]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "M[3]U[4]S[5]    K[502]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "M"),
                                Token("CONDITION_EXPRESSION", "[3]U[4]"),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "S"),
                                Token("CONDITION_EXPRESSION", "[5]    "),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "K"),
                                Token("CONDITION_EXPRESSION", "[502]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "m[3]u[4]s[5]    k[502]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "m"),
                                Token("CONDITION_EXPRESSION", "[3]u[4]"),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "s"),
                                Token("CONDITION_EXPRESSION", "[5]    "),
                            ],
                        ),
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("MODAL_MARK", "k"),
                                Token("CONDITION_EXPRESSION", "[502]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "U[1]O[5]",
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "U"),
                                Token("CONDITION_EXPRESSION", "[1]O[5]"),
                            ],
                        ),
                    ],
                ),
            ),
            pytest.param(
                "u[1]O[5]",  # lower case "u"
                Tree(
                    "ahb_expression",
                    [
                        Tree(
                            "single_requirement_indicator_expression",
                            [
                                Token("PREFIX_OPERATOR", "u"),
                                Token("CONDITION_EXPRESSION", "[1]O[5]"),
                            ],
                        ),
                    ],
                ),
            ),
        ],
    )
    def test_parse_valid_ahb_expression_to_to_single_requirement_indicator_expressions(
        self, ahb_expression: str, expected_tree: Tree[Token]
    ):
        """
        Tests that valid ahb expressions are parsed as expected.
        """
        parsed_tree = parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)

        assert isinstance(parsed_tree, Tree)
        assert parsed_tree == expected_tree

    @pytest.mark.parametrize(
        "ahb_expression",
        [
            pytest.param(None),
            pytest.param(""),
            pytest.param("   "),
            pytest.param("[1]"),  # without requirement indicator
            pytest.param("Mus[2]"),  # this is a tricky one because since we introduced lower case operators and modal
            # marks the characters alone are all ok (lower case 'u' is a valid expression operator and lower case 's'
            # could be interpreted as lower case model mark). As of 2021-10-27, when parsing an AHB expression we only
            # use a simple regex to roughly check if the characters at the positions where we expect condition
            # expressions to be found are allow-listed. But we do not check the grammar yet. The grammar is checked
            # upon parsing the condition expression, but that's the second step. To keep the test failing, I introduced
            # a negative look ahead where formerly only a "character set" allow list check.
            pytest.param("Muss[2]C[3]"),
        ],
    )
    def test_parse_invalid_ahb_expression(self, ahb_expression: str):
        """Tests that an error is raised when trying to parse an invalid string."""

        with pytest.raises(SyntaxError) as excinfo:
            parse_ahb_expression_to_single_requirement_indicator_expressions(ahb_expression)

        assert """Please make sure that the ahb_expression starts with a requirement indicator \
(i.e Muss/M, Soll/S, Kann/K, X, O, U) and the condition expressions consist of only \
the following characters: [ ] ( ) U ∧ O ∨ X ⊻ and digits.""" in str(
            excinfo.value
        )



"""
Enums used in AHB and condition expressions.
"""
from enum import Enum, unique
from typing import Dict, Literal, Union

from marshmallow import Schema, fields, post_dump, post_load, pre_load


@unique
class ModalMark(str, Enum):
    """
    A modal mark describes if information are obligatory or not. The German term is "Merkmal".
    The modal marks are defined by the EDI Energy group (see edi-energy.de → Dokumente → Allgemeine Festlegungen).
    The modal mark stands alone or before a condition expression.
    It can be the start of several requirement indicator expressions in one AHB expression.
    """

    MUSS = "MUSS"
    """
    German term for "Must". Is required for the correct structure of the message.
    If the following condition is not fulfilled, the information must not be given ("must not")
    """

    SOLL = "SOLL"
    """
    German term for "Should". Is required for technical reasons.
    Always followed by a condition.
    If the following condition is not fulfilled, the information must not be given.
    """

    KANN = "KANN"
    """
    German term for "Can". Optional
    """


@unique
class PrefixOperator(str, Enum):
    """
    Operator which does not function to combine conditions, but as requirement indicator.
    It stands alone or in front of a condition expression. Please find detailed descriptions of the operators and their
    usage in the "Allgemeine Festlegungen".
    Note that with MaKo2022 introced 2022-04-01 the "O" and "U" prefix operators will be deprecated.
    Refer to the "Allgemeine Festlegungen" valid up to 2022-04-01 for deprecated "O" and "U".
    """

    X = "X"
    """
    The "X" operator. See "Allgemeine Festlegungen" Kapitel 6.8.1. Usually this just means something is required
    or required under circumstances defined in a trailing condition expression.
    It shall be read as "exclusive or" regarding how qualifiers/codes shall be used from a finite set.
    Note that "X" can also be used as "logical exclusive or" (aka "xor") operator in condition expressions.
    The prefix operator works differently from the logical operator in condition expressions!
    The usage of "X" as logical operator is deprecated since 2022-04-01. It will be replaced with the "⊻" symbol.
    """
    O = "O"
    """
    The "O" operator means that at least one out of multiple possible qualifiers/codes has to be given.
    This is typically found when describing ways to contact a market partner (CTA): You can use email or phone or fax
    but you have to provide at least one of the given possibilities.
    The usage of "O" as a prefix operator is deprecated since 2022-04-01.
    Note that "O" can also be used as a "logical or" (aka "lor") operator in condition expressions.
    The prefix operator works differently from the logical operator in condition expressions!
    The usage of "O" as logical operator is also deprecated since 2022-04-01. It will be replaced with the "∨" symbol.
    """
    U = "U"
    """
    The "U" operator means that all provided qualifiers/codes have to be used.
    The usage of "U" as a prefix operator is deprecated since 2022-04-01.
    Note that "U" can also be used as a "logical and" (aka "land") operator in condition expressions.
    The prefix operator works differently from the logical operator in condition expressions!
    The usage of "U" as logical operator is also deprecated since 2022-04-01. It will be replaced with the "∧" symbol.
    """


RequirementIndicator = Union[PrefixOperator, ModalMark]
"""
A Requirement Indicator is either the Merkmal :class:`ModalMark` or the :class:`PrefixOperator` of the
data element/data element group/segment/segment group at which it is used.
"""


# pylint:disable=no-self-use, unused-argument
class RequirementIndicatorSchema(Schema):
    """
    a helper schema because marshmallow does not support something like fields.Union out of the box
    """

    value = fields.String()

    @pre_load
    def pre_load(self, data, **kwargs) -> Dict[Literal["value"], str]:
        """puts the value in an artificial dictionary"""
        return {"value": data}

    @post_load
    def post_load(self, data, **kwargs) -> RequirementIndicator:
        """tries to parse the data as either PrefixOperator or ModalMark"""
        try:
            return ModalMark(data["value"])
        except ValueError:
            return PrefixOperator(data["value"])

    @post_dump
    def post_dump(self, data, **kwargs):
        """
        returns the enum value as upper case
        """
        return data["value"].upper()


class LogicalOperator(str, Enum):
    """
    Logical operators connect two tokens.
    """

    LAND = "U"
    """
    logical AND, also denoted as "∧", used in and_composition
    """

    LOR = "O"
    """
    logical OR, also denoted as "∨", used in or_composition
    """

    XOR = "X"
    """
    logical excluxive OR (XOR), also denoted as "⊻", used in xor_composition
    """



""" Tests the dictionary based RC evaluator"""
from unittest import mock

import pytest  # type:ignore[import]

from ahbicht.content_evaluation.rc_evaluators import DictBasedRcEvaluator
from ahbicht.expressions.condition_nodes import ConditionFulfilledValue

pytestmark = pytest.mark.asyncio


class TestDictBasedRcEvaluator:
    """Test for the evaluation using the Dict Based RC Evaluator"""

    async def test_evaluation(self, mocker):
        hardcoded_results = {
            "1": ConditionFulfilledValue.NEUTRAL,
            "2": ConditionFulfilledValue.UNFULFILLED,
            "3": ConditionFulfilledValue.FULFILLED,
            "4": ConditionFulfilledValue.UNKNOWN,
        }
        evaluator = DictBasedRcEvaluator(hardcoded_results)
        assert await evaluator.evaluate_single_condition("1") == ConditionFulfilledValue.NEUTRAL
        assert await evaluator.evaluate_single_condition("2") == ConditionFulfilledValue.UNFULFILLED
        assert await evaluator.evaluate_single_condition("3") == ConditionFulfilledValue.FULFILLED
        assert await evaluator.evaluate_single_condition("4") == ConditionFulfilledValue.UNKNOWN
        with pytest.raises(NotImplementedError):
            await evaluator.evaluate_single_condition("5")

        single_condition_spy = mocker.spy(evaluator, "evaluate_single_condition")
        assert await evaluator.evaluate_conditions(["1", "2", "3", "4"]) == hardcoded_results
        single_condition_spy.assert_has_awaits(
            [mock.call("1", None), mock.call("2", None), mock.call("3", None), mock.call("4", None)]
        )



"""
Dataclasses that are relevant in the context of the content_evaluation.
"""
import dataclasses
from dataclasses import dataclass
from typing import Optional, Union


@dataclass
class EvaluatableData:
    """
    Data that can be processed/evaluated by an evaluator. They must not change during a content_evaluation run.
    The evaluatable data act as a flexible container to pass information that might be necessary for the
    content_evaluation. As of now (because our content_evaluation capabilities are quite limited) the only information
    provided is the meta seed of the message itself. But in the future the data provided might grow.
    """

    edifact_seed: Union[dict, list]  # the meta seed of the message that is being validated
    # ideas for what else could go here:
    # - pruefidentifikator to tweak the content_evaluation depending on the situation?


@dataclass
class EvaluationContext:
    """
    A content_evaluation context describes the setting in which a condition shall be evaluated. The content_evaluation
    context might have different values for the same condition in one content_evaluation run. E.g. if the purpose of the
    condition is to make sure that every Zähler with zähler type "EHZ" has some properties the context of the
    content_evaluation is one zähler entry although there might be multiple zählers present in the message.
    """

    scope: Optional[
        str
    ]  # jsonpath that refers to the scope of the content_eval. If None, then "$" = entire message is used as scope.


def copy_evaluation_context(context: EvaluationContext) -> EvaluationContext:
    """
    Returns a deep copy of the provided context.
    This allows you to create a copy of a context instead of modifying the original context (as EvaluationContexts are
    "pass by reference")
    :param context:
    :return: a deep copy of the context
    """
    return dataclasses.replace(context)
